package model;

import java.awt.Color;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Scanner;

public class GameBoard {
    public final int scale = 24;
    public final int playerSpeed = 5;
    public final int turnTolerance = playerSpeed;
    public final int representationSize;
    protected final ArrayList<Player> players = new ArrayList<>();

    private int width, height;
    private Item origTable[][];
    public  Item table[][];
    private int numFoods, numFoodsRemain;

    public GameBoard(){
        loadGame();
        initGame();
        representationSize = players.size() * Player.representationSize + (width * height + 7) / 8;
    }
    
    private void loadGame(){
        ArrayList<String> lines = new ArrayList<>();
        int maxWidth = 0;
        try {
            InputStream is = getClass().getResourceAsStream("../res/pacman.txt");
            Scanner s = new Scanner(is);
            while (s.hasNextLine()){
                String line = s.nextLine();
                lines.add(line);
                if (line.length() > maxWidth) maxWidth = line.length();
            }
            is.close();
        } catch (IOException ex) {}
        width = maxWidth;
        height = lines.size();
        numFoods = 0;
        
        if (width > 0 && height > 0){
            origTable = new Item[height][width];
            table = new Item[height][width];
            for (int y = 0; y < height; y++){
                String line = lines.get(y);
                for (int x = 0; x < line.length(); x++){
                    char c = line.charAt(x);
                    if (c == ' '){
                        origTable[y][x] = Item.EMPTY;
                    } else if (c == '.'){
                        origTable[y][x] = Item.FOOD;
                        numFoods++;
                    } else if (c == 'o'){
                        origTable[y][x] = Item.POWER;
                    } else if (c == 'X'){
                        origTable[y][x] = Item.WALL;                        
                    } else if (c == '*'){
                        players.add(new Player(x, y, scale, Color.YELLOW, playerSpeed));
                    } else {
                        origTable[y][x] = Item.EMPTY;
                    }
                }
            }
        }
    }
    
    public final void initGame(){
        for (int y = 0; y < height; y++){
            for (int x = 0; x < width; x++){
                table[y][x] = origTable[y][x];
            }
        }
        numFoodsRemain = numFoods;
        for (int i = 0; i < players.size(); i++) {
            players.get(i).reset();
        }
    }

    public int getWidth() { return width; }

    public int getHeight() { return height; }

    // Does not need synchronization, because it never changes.
    public final int numPlayers(){ return players.size(); }

    // Player fields can be read and written by multiple threads.
    public synchronized Player getPlayer(int id){ return new Player(players.get(id)); }
    
    // Table changes each time a PacMan eats something.
    public synchronized Item[][] getTable(){
        Item copy[][] = new Item[height][width];
        for (int y = 0; y < height; y++){
            for (int x = 0; x < width; x++){
                copy[y][x] = table[y][x];
            }
        }
        return copy;
    }
        
    // This method is called only by a single application, or a single server thread (exclusively)
    // Clients call the overriden version, which sends the direction to the server
    public void changePlayerDirection(int player, Direction d){
        Player p = players.get(player);
        boolean sameAxis = (p.moveDir.isHorizontal == d.isHorizontal && p.moveDir.isVertical == d.isVertical);
        if (sameAxis || p.moveDir == Direction.STAY){
            // leave or reverse the direction, or start the staying pacman
            p.nextDir = p.moveDir = p.faceDir = d;
        } else {
            // Turn pacman
            p.nextDir = d;
        }
    }
    
    // This method is called only by a single application, or a single server thread (exclusively)
    // Clients never call movePlayers
    public void movePlayers() {
        for (int i = 0; i < players.size(); i++){
            Player p = players.get(i);
            Direction m = p.moveDir;
            Direction n = p.nextDir;
            if (m == Direction.STAY) continue;
            int offsX = 0, offsY = 0;
            int predX = p.smoothPos.x + m.x * playerSpeed;
            int predY = p.smoothPos.y + m.y * playerSpeed;

            boolean hasToTurn = (m.isHorizontal == n.isVertical);
            boolean canTurnFromDown = (p.smoothPos.y - p.scaledPos.y <= turnTolerance) && (table[p.pos.y][p.pos.x + n.x] != Item.WALL);
            boolean canTurnFromRight = (p.smoothPos.x - p.scaledPos.x <= turnTolerance) && (table[p.pos.y + n.y][p.pos.x] != Item.WALL);
            boolean canTurnFromUp = (p.scaledPos.y + scale - p.smoothPos.y <= turnTolerance) && (table[p.pos.y + 1][p.pos.x + n.x] != Item.WALL);
            boolean canTurnFromLeft = (p.scaledPos.x + scale - p.smoothPos.x <= turnTolerance) && (table[p.pos.y + n.y][p.pos.x + 1] != Item.WALL);

            if (hasToTurn){
                hasToTurn = false;
                if (m == Direction.RIGHT && canTurnFromLeft){
                    hasToTurn = true;
                    p.setSmoothX(p.scaledPos.x + scale);
                } else if (m == Direction.LEFT && canTurnFromRight){
                    hasToTurn = true;
                    p.setSmoothX(p.scaledPos.x);
                } else if (m == Direction.DOWN && canTurnFromUp){
                    hasToTurn = true;
                    p.setSmoothY(p.scaledPos.y + scale);
                } else if (m == Direction.UP && canTurnFromDown){
                    hasToTurn = true;
                    p.setSmoothY(p.scaledPos.y);
                }
                if (hasToTurn) {
                    m = p.faceDir = p.moveDir = p.nextDir;
                    tryEat(p.pos.x, p.pos.y);
                }
            }

            if (m == Direction.RIGHT){
                int indRight = (predX + scale - 1) / scale;
                int distRight = (indRight - 1) * scale - p.smoothPos.x;
                boolean canMoveRight = table[p.pos.y][indRight] != Item.WALL;
                offsX = (canMoveRight) ? playerSpeed : distRight;
            } else if (m == Direction.LEFT){
                int indLeft = predX / scale;
                int distLeft = p.smoothPos.x - p.scaledPos.x;
                boolean canMoveLeft = table[p.pos.y][indLeft] != Item.WALL;
                offsX = (canMoveLeft) ? -playerSpeed : -distLeft;
            } else if (m == Direction.UP){
                int indUp = predY / scale;
                int distUp = p.smoothPos.y - p.scaledPos.y;
                boolean canMoveUp = table[indUp][p.pos.x] != Item.WALL;
                offsY = (canMoveUp) ? -playerSpeed : -distUp;
            } else if (m == Direction.DOWN){
                int indDown = (predY + scale - 1) / scale;
                int distDown = (indDown - 1) * scale - p.smoothPos.y;
                boolean canMoveDown = table[indDown][p.pos.x] != Item.WALL;
                offsY = (canMoveDown) ? playerSpeed : distDown;
            }

            if (offsX != 0 || offsY != 0){
                p.animateMouth();
                p.setSmoothCoord(p.smoothPos.x + offsX, p.smoothPos.y + offsY);
                tryEat(p.pos.x, p.pos.y);
            } else {
                p.moveDir = Direction.STAY;
            }
        }
        if (numFoodsRemain == 0){
            initGame();
        }
    }
    
    // Called by movePlayers() to check whether PacMan can eat or not
    private void tryEat(int x, int y){
        if (table[y][x] == Item.FOOD) numFoodsRemain--;
        table[y][x] = Item.EMPTY;
    }
    
    // Called only by the server thread
    public byte[] toByteArray() {
        // The representation of the classic PacMan level requires 1 + n x 9 + 109 bytes
        // May write instead into a DataOutputStream(new ByteArrayOutputStream())
        int numBytes = representationSize;
        int byteIndex = 0;
        byte byteValue = 0;
        byte bitIndex = 0;
        ByteBuffer bb = ByteBuffer.allocate(numBytes);
        for (int i = 0; i < players.size(); i++){
            bb.put(players.get(i).toByteArray());
        }
        for (int y = 0; y < height; y++){
            for (int x = 0; x < width; x++){
                byteValue <<= 1; // shift 1 bit left (multiply by 2)
                byteValue |= (table[y][x] != origTable[y][x]) ? 1 : 0; // mark the differences
                if (bitIndex++ == 7) { // put the completed byte into the array
                    bb.put(byteValue);
                    bitIndex = 0;
                    byteIndex++;
                }
            }
        }
        if (numBytes > byteIndex) {
            byteValue <<= 8 - bitIndex; // shift with padding bits
            bb.put(byteValue);
        }
        return bb.array();
    }
    
    // Called by the client or the constructor
    // It updates the GameBoard model, including the players
    // These data can be read at anytime by the Client's GUI
    public final synchronized void fromByteArray(byte[] array){
        if (array == null || array.length != representationSize) return;
        ByteArrayInputStream bais = new ByteArrayInputStream(array);
               
        try {
            for (int i = 0; i < players.size(); i++){
                players.get(i).updateFromStream(bais);
            }
            int byteIndex = players.size() * Player.representationSize;
            int bitValue = 128; // start with the Most Significant Bit (MSB)
            for (int y = 0; y < height; y++){
                for (int x = 0; x < width; x++){
                    if ((array[byteIndex] & bitValue) > 0){ // select the bit
                        table[y][x] = Item.EMPTY;
                    }
                    bitValue >>= 1; // go to the next bit in MSB->LSB direction
                    if (bitValue == 0){
                        bitValue = 128;
                        byteIndex++;
                    }
                }
            }
        } catch (IOException e){}
    }
}
