package view;

import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.URL;
import javax.swing.*;
import model.Direction;
import model.GameBoard;
import networking.*;

public class GameWindow extends JFrame implements KeyListener, ServerDisplay, ClientDisplay {
    private final GameBoard localGame = new GameBoard();
    private final GameBoardNetServer server;    
    private final GameBoardNetClient client;
    private final GameChooserPanel gameChooserPanel = new GameChooserPanel();
    private final ProgressDialog progressDialog = new ProgressDialog(this, "Starting", "Waiting for guest");
    private GameMode gameMode = GameMode.LOCAL;

    private final Timer timer = new Timer(20, e -> {
        localGame.movePlayers();
        refresh();
    });
    
    /**
     * Creates new form GameWindow
     */
    public GameWindow() throws IOException {
        
        server = new GameBoardNetServer(25, 5000);
        client = new GameBoardNetClient();
        server.setDisplay(this);
        timer.setRepeats(true);

        initComponents();
        URL url = GameWindow.class.getClassLoader().getResource("res/pacman.png");
        setIconImage(Toolkit.getDefaultToolkit().getImage(url));
        setLocationRelativeTo(null);
        addKeyListener(this);
        
        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent we) {
                closeGame();
            }
        });
        
        gamePanel.setGameBoard(localGame);
        
        pack();

        setVisible(true);
        startGame();
    }
    
    private void startGame(){
        GameChooserPanel gcp = gameChooserPanel;
        stopGame();
        int ret = JOptionPane.showConfirmDialog(null, gcp, "Select game type", JOptionPane.OK_CANCEL_OPTION);
        if (ret != JOptionPane.OK_OPTION) {
            System.exit(0);
        }

        gameMode = gcp.getGameMode();
        
        if (gameMode == GameMode.LOCAL){
            setTitle("PacMan - Two players!");
            localGame.initGame();
            gamePanel.setGameBoard(localGame);
            timer.start();
        } else if (gameMode == GameMode.SERVER){
            setTitle("PacMan - Multiplayer (host)");
            server.initGame();
            client.initGame();
            gamePanel.setGameBoard(client);
            server.start();
            client.setDisplay(this);
            client.setServerAddress("127.0.0.1", gcp.getPort());
            client.start();
            int reason = progressDialog.showDialog(0);
            System.out.println("Reason: " + reason);
        } else {
            setTitle("PacMan - Multiplayer (guest)");
            client.initGame();
            gamePanel.setGameBoard(client);
            client.setDisplay(this);
            client.setServerAddress(gcp.getAddress(), gcp.getPort());
            client.start();
        }
    }
    
    private void stopGame(){
        if (gameMode == GameMode.LOCAL){
            timer.stop();
        } else if (gameMode == GameMode.SERVER){
            server.stop();
            client.stop();
        } else {
            client.stop();
        }        
    }
    
    private void closeGame(){
        stopGame();
        System.exit(0);
    }

    /**
     * This method is called from within the constructor to initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is always
     * regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        gamePanel = new GamePanel();
        jMenuBar1 = new javax.swing.JMenuBar();
        jMenu1 = new javax.swing.JMenu();
        menuItemNewGame = new javax.swing.JMenuItem();
        jSeparator1 = new javax.swing.JPopupMenu.Separator();
        menuItemExit = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle("PacMan");
        setResizable(false);

        gamePanel.setBackground(new java.awt.Color(255, 255, 255));

        javax.swing.GroupLayout gamePanelLayout = new javax.swing.GroupLayout(gamePanel);
        gamePanel.setLayout(gamePanelLayout);
        gamePanelLayout.setHorizontalGroup(
            gamePanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 833, Short.MAX_VALUE)
        );
        gamePanelLayout.setVerticalGroup(
            gamePanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 333, Short.MAX_VALUE)
        );

        jMenu1.setText("Game");

        menuItemNewGame.setText("New Game");
        menuItemNewGame.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuItemNewGameActionPerformed(evt);
            }
        });
        jMenu1.add(menuItemNewGame);
        jMenu1.add(jSeparator1);

        menuItemExit.setText("Exit");
        menuItemExit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                menuItemExitActionPerformed(evt);
            }
        });
        jMenu1.add(menuItemExit);

        jMenuBar1.add(jMenu1);

        setJMenuBar(jMenuBar1);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(gamePanel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(gamePanel, javax.swing.GroupLayout.DEFAULT_SIZE, 333, Short.MAX_VALUE)
                .addContainerGap())
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void menuItemExitActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuItemExitActionPerformed
        closeGame();
    }//GEN-LAST:event_menuItemExitActionPerformed

    private void menuItemNewGameActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_menuItemNewGameActionPerformed
        startGame();
    }//GEN-LAST:event_menuItemNewGameActionPerformed

    /**
     * @param args the command line arguments
     */
    public static void main(String args[]) {
        /* Set the Nimbus look and feel */
        //<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
         * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html 
         */
        try {
            for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
                if ("Windows".equals(info.getName())) {
                    javax.swing.UIManager.setLookAndFeel(info.getClassName());
                    break;
                }
            }
        } catch (ClassNotFoundException ex) {
            java.util.logging.Logger.getLogger(GameWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (InstantiationException ex) {
            java.util.logging.Logger.getLogger(GameWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (IllegalAccessException ex) {
            java.util.logging.Logger.getLogger(GameWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (javax.swing.UnsupportedLookAndFeelException ex) {
            java.util.logging.Logger.getLogger(GameWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        }
        //</editor-fold>

        /* Create and display the form */
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                try {
                    new GameWindow().setVisible(true);
                } catch (IOException ex) {}
            }
        });
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private view.GamePanel gamePanel;
    private javax.swing.JMenu jMenu1;
    private javax.swing.JMenuBar jMenuBar1;
    private javax.swing.JPopupMenu.Separator jSeparator1;
    private javax.swing.JMenuItem menuItemExit;
    private javax.swing.JMenuItem menuItemNewGame;
    // End of variables declaration//GEN-END:variables

    

    @Override
    public void keyTyped(KeyEvent ke) {
    }

    @Override
    public void keyPressed(KeyEvent ke) {
        GameMode mode = GameMode.LOCAL;
        Direction d = Direction.STAY;

        if (ke.getKeyCode() == KeyEvent.VK_A){
            mode = GameMode.SERVER;
            d = Direction.LEFT;
        } else if (ke.getKeyCode() == KeyEvent.VK_D){
            mode = GameMode.SERVER;
            d = Direction.RIGHT;
        } else if (ke.getKeyCode() == KeyEvent.VK_W){
            mode = GameMode.SERVER;
            d = Direction.UP;
        } else if (ke.getKeyCode() == KeyEvent.VK_S){
            mode = GameMode.SERVER;
            d = Direction.DOWN;
        } else if (ke.getKeyCode() == KeyEvent.VK_LEFT){
            mode = GameMode.CLIENT;
            d = Direction.LEFT;
        } else if (ke.getKeyCode() == KeyEvent.VK_RIGHT){
            mode = GameMode.CLIENT;
            d = Direction.RIGHT;
        } else if (ke.getKeyCode() == KeyEvent.VK_UP){
            mode = GameMode.CLIENT;
            d = Direction.UP;
        } else if (ke.getKeyCode() == KeyEvent.VK_DOWN){
            mode = GameMode.CLIENT;
            d = Direction.DOWN;
        }

        if (gameMode == GameMode.LOCAL){
            localGame.changePlayerDirection((mode == GameMode.SERVER) ? 0 : 1, d);
        } else if (mode == gameMode){
            client.changePlayerDirection(d);
        }
    }

    @Override
    public void keyReleased(KeyEvent ke) {}

    // ------------------------------------------------------------------------
    
    @Override
    public void refresh() {
        gamePanel.repaint();
    }

    @Override
    public void connectionLost(int client) {
        if (client == 1){
            JOptionPane.showMessageDialog(this, "Connection lost to remote client", "Connection error", JOptionPane.ERROR_MESSAGE);
        }
    }

    @Override
    public void connectionLost() {
        if (gameMode == GameMode.CLIENT){
            JOptionPane.showMessageDialog(this, "Connection lost to server", "Connection error", JOptionPane.ERROR_MESSAGE);
        }
    }

    @Override
    public void cannotOpenServer() {
        JOptionPane.showMessageDialog(this, "Cannot open port", "Connection error", JOptionPane.ERROR_MESSAGE);
    }

    @Override
    public void cannotConnect() {
        JOptionPane.showMessageDialog(this, "Cannot connect to the server", "Connection error", JOptionPane.ERROR_MESSAGE);
    }

    @Override
    public void allClientsAreConnected() {
        progressDialog.hideDialog();
    }
}
